"""
Runtime assertion utility.

Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""


def runtime_assert(condition: bool, message: str = "") -> None:
    """Raises RuntimeError if condition is False.

    Unlike assert statements, this is never stripped by the Python optimizer (-O flag)
    and does not trigger Bandit B101 warnings.
    """
    if not condition:
        raise RuntimeError(message)
