@import "dark_colours.css";
@import "dark_buttons.css";
@import "connection_status.css";
@import "main_widget.css";
@import "quick_connect_widget.css";
@import "search_entry.css";
@import "search_results.css";
@import "server_list.css";
@import "nps_survey_modal.css";

entry:focus-within,
button:focus-visible,
label:focus-visible,
label link:focus-visible,
switch:focus-visible {
    outline-color: @interaction-norm;
}

entry:focus-within {
    border-color: @interaction-norm;
    box-shadow: none;
}

switch {
    background-color: @field-norm;
    background-image: image(@field-norm);
    border-color: @field-norm;
    border-radius: 999px;
}

switch:checked {
    background-color: @interaction-norm;
    background-image: image(@interaction-norm);
    border-color: @interaction-norm;
}

switch:checked:hover {
    background-color: @interaction-norm-hover;
    background-image: image(@interaction-norm-hover);
    border-color: @interaction-norm-hover;
}

switch slider {
    border-radius: 999px;
    background-color: @text-norm;
    background-image: image(@text-norm);
}

checkbutton > check:checked,
checkbutton > check:indeterminate {
    background-color: @interaction-norm;
    background-image: image(@interaction-norm);
    border-color: @interaction-norm;
    color: @text-norm;
}

checkbutton > check:checked:hover,
checkbutton > check:indeterminate:hover {
    background-color: @interaction-norm-hover;
    background-image: image(@interaction-norm-hover);
    border-color: @interaction-norm-hover;
}

checkbutton > radio {
    border-radius: 999px;
}

checkbutton > radio:checked {
    background-color: @interaction-norm;
    background-image: image(@interaction-norm);
    border-color: @interaction-norm;
    color: @text-norm;
}

entry text selection {
    background-color: @interaction-norm;
    color: @text-norm;
}

label link {
    color: @interaction-norm;
}

label link:hover {
    color: @interaction-norm-hover;
}

scrollbar {
    border: none;
}

scrollbar slider,
scrollbar slider:hover {
    background-color: @border-norm;
}

label.info {
    background-color: @signal-info;
    background-image: image(@signal-info);
    color: @text-norm;
    text-shadow: None;
    padding: 5px;
}

label.error {
    background-color: @signal-danger;
    background-image: image(@signal-danger);
    color: @text-norm;
    text-shadow: None;
    padding: 5px;
}

label.success {
    background-color: @signal-success;
    background-image: image(@signal-success);
    color: @text-norm;
    text-shadow: None;
    padding: 5px;
}

#login-logo {
    margin-top: 75px;
    margin-bottom: 40px;
    margin-right: 10px;
}

#two-factor-auth-vpn-logo {
    margin-top: 2em;
    margin-bottom: 1em;
}

#bug-report-content,
#bug-report-dialog buttonbox,
#early-access-dialog buttonbox,
.dialog-action-area{
    margin: 20px;
}
.dialog-action-area > button {
    margin-left: 5px;
}

.setting-category {
    margin-top: 20px;
    margin-left: 20px;
    margin-right: 20px;
}

.setting-category:last-child {
    margin-bottom: 60px;
}

.setting-item {
    border: 1px solid rgba(155, 155, 155, 0.2);
    padding: 10px;
    border-radius: 10px;
}

.upgrade-tag {
    border-radius: 10px;
    background: transparent;
    border: 1px solid rgba(155, 155, 155, 0.2);
    outline: none;
    background-image: none;
}

.beta-tag {
    border-radius: 5px;
    border: 1px solid @primary;
    color: @primary;
    padding: 2px 6px;
    font-size: 0.9em;
}

.release-notes-collection {
    margin: 20px 20px 60px 20px;
}

#login-kill-switch-revealer box {
    border-top: 1px solid rgba(155, 155, 155, 0.2);
    padding: 15px 0px;
    margin-top: 20px;
}

label.signal-danger {
    color: @signal-danger;
}

#split-tunneling-app-selection-window viewport {
    margin: 15px 30px 30px;
    padding: 2px 0px 0px 0px;
    border: 1px solid @border-norm;
    border-radius: 5px;
}

#split-tunneling-app-selection-window viewport > box {
    margin: 0px 0px 0px 5px;
}

#split-tunneling-app-selection-window #split-tunneling-app-done-button {
    margin: 15px 30px 30px;
}

#split-tunneling-app-based-settings {
    margin-top: 10px;
}

#split-tunneling-app-based-settings #split-tunneling-app-add-button {
    margin: 15px 10px 5px 10px;
}


#selected-app-list {
    padding: 12px 16px 12px 16px;
    border-radius: 5px;
    border: 1px solid rgba(155, 155, 155, 0.2);
}

#two-factor-auth-widget > label.two-factor-auth-stack-title {
    font-size: 1.3em;
    font-weight: bold;
    margin-bottom: 0.5em;
}

.default-loading-widget-label{
    margin-left: 1.8em;
	margin-right: 1.8em;
}

