@import "dark_colours.css";

button.primary:not(:disabled) {
	background-color: @interaction-norm;
	background-image: image(@interaction-norm);
	border-color: @interaction-norm;
    color: @text-norm;
	text-shadow: none;
}

button.secondary:hover, button.primary:hover,
button.secondary:focus, button.primary:focus {
	background-color: @interaction-norm-hover;
	background-image: image(@interaction-norm-hover);
	border-color: @interaction-norm-hover;
	color: @text-norm;
	text-shadow: none;
}

button.danger:hover,
button.danger:focus {
	border-color: @signal-danger;
    background-color: @signal-danger;
	background-image: image(@signal-danger);
	color: @text-norm;
	text-shadow: none;
}

button.secondary:active, button.primary:active {
	background-image: image(@interaction-norm-active);
	border-color: @interaction-norm-active;
	color: @text-norm;
	text-shadow: none;
}

button.danger:active {
	border-color: @signal-danger-active;
    background-color: @signal-danger-active;
	background-image: image(@signal-danger-active);
	color: @text-norm;
	text-shadow: none;
}

button.destructive-action:not(:disabled) {
	background-color: @signal-danger;
	background-image: image(@signal-danger);
	border-color: @signal-danger;
	color: @text-norm;
	text-shadow: none;
}

button.destructive-action:not(:disabled):hover,
button.destructive-action:not(:disabled):focus {
	background-color: @signal-danger-hover;
	background-image: image(@signal-danger-hover);
	border-color: @signal-danger-hover;
	color: @text-norm;
	text-shadow: none;
}

button.destructive-action:not(:disabled):active {
	background-color: @signal-danger-active;
	background-image: image(@signal-danger-active);
	border-color: @signal-danger-active;
	color: @text-norm;
	text-shadow: none;
}

button.spaced label {
	padding: 5px 20px;
}

button.link label {
    color: @interaction-norm;
}

